

#-----------------         NOTES            -----------------------------------
#
# 1. if a cue value is empty, the cue has to be copied from the goal buffer slot
#   >> make this more generic
#
#
# 2. ToDo's
#	**	. activation boost shd occur only after =retrieval.. currently it's for every retrieval eg. picture retrieval in aphasia model
#	**	. if a chunk doesn't have specification for certain feature then that value (which is 'none') is not considered in mismatch-penalty
#		. create one list for each buffer and each side (lhs/rhs) of the productions (eg. goal.lhs, goal.rhs, NPb.lhs (??),
#		  NPb.rhs (chunk mmodification), NPb.create (new chunk) etc.)
#		
#		. write a fuction which can flush out current DM contents
#		. document the entire procedure/algorithm followed in the package for future enhancement/debugging
#		. if a 'stage' doesn't have a new chunk creation then the 'attach-*' prod is not fired and no addition of 50 ms cycle time
#		. implement retrieval threshold logic (retrieval failure)       [DONE]
#		. add production utility       [DONE]
#		. add goal buffer modification table        [DONE]
#		. add goal buffer activation spreading (??)
#		. possibility of making a chunk slot contain another chunk as its value
#		. creation of multiple chunks in one stage  [DONE]
#		. replace all 'for' loops with 'while' loops
#		. try utility eq. from act-r 5 (why?)
#
# 3. REMs (limitations):
#		. the system doesn't always have the information of currently what is the content of a buffer (hence no procedure of 'chunk merging')
#		. slot values can not be chunks in a buffer. e.g. comp = +VPb
#		. can not modify contents of chunks in syntactic buffers (VPb, NPb etc.) w/o retrieving them
#		. various chunk slot values like 'nom' are not chunks themselves, but just strings
#		. 1st entry in the retrieval list elements should always be a 'type' feature
#		. 1st 3 entries for every chunk in the DM should be 'type', 'ctime' & 'npres' (not anymore!)
#		. In a typical model the 'W' value of activation spread is high due to slots like 'goal-cat', 'clause' etc. in the goal buffer
#		. can not have different utility values for 'Retrieval-request' and 'Attachment' productions from the same stage
#
#-------------------------------------------------------------------------------





## -- Functions --


######################################################
# adds new chunk to the DM
######################################################

add.chunk <- function(newChunk){
	
	if(verbose){
	print("adding chunk:", quote=FALSE)
	print(paste("", newChunk), quote=FALSE)
	}
	
	chunkID <<- chunkID+1
	chunks[[chunkID,1]] <<- newChunk
	ret.hist[chunkID,1] <<- currTime
	}


######################################################
# increment the time counter
######################################################

add.time <- function(dur){
	currTime <<- currTime + dur
	
	if(verbose){
	print(paste("Time: ", currTime), quote=FALSE)
	}
	}


######################################################
# conflict resolution: find the production to fire
######################################################

conflict.resolution <- function(stageNo){
	if(verbose.func){
		print("in conflict.resolution", quote=FALSE)
		}
		
	prods = array(, c(steps,2))
	
	i = currStep+1
	while(i <= steps){
		if(util.steps[i,1] == stageNo){
			if(util.steps[i,2] == 0){
				stepNo <<- i
				return(stepNo)
				}
			else{
				j=i
				k=1
				while(util.steps[j,1] == stageNo){
					prods[k,1] = j
					prods[k,2] = util.steps[j,2] + noise(0, par.egs)
					j=j+1
					k=k+1
					}
				stepNo = cbind(prods[order(prods[,2], decreasing=TRUE),1], prods[order(prods[,2], decreasing=TRUE),2])[1,1]
				return(stepNo)
				
				}
			}
		i=i+1
		}
print("THIS SHOULD NOT BE PRINTED returning stepNo", quote=FALSE)
	
	return(stepNo)
	}



######################################################
# checks if any change in the goal buffer is requested
######################################################

check.goal <- function(){
	nm = names(goal.steps[[currStep,1]])
	
	if(!is.null(nm)){
		
		j=1
		while(j <= length(nm)){
			goal.chunk[[nm[j]]] <<- goal.steps[[currStep,1]][[nm[j]]]
			j=j+1
			}
		}
	
	}



######################################################
# prepares retrieval que according to the 
# retrieval request
######################################################

retrieval.request <- function(cues){
	if(verbose.func){
		print("in retrieval.request", quote=FALSE)
		}
		
	list.cues(cues)
	build.retQ()

	if(verbose){
	print(cues.list, quote=FALSE)

	print("",quote=FALSE)
	print("                 Activation Spreading", quote=FALSE)
	}
	
	compute.act.spreading()
	compute.base.act()

	if(verbose){
	print("",quote=FALSE)
	print("                 Partial Match", quote=FALSE)
	}

	compute.partial.match()
	sort.retQ()		#! sort the retQ on ret. time
	}



######################################################
# performs actual retrieval
######################################################

retrieve <- function(){
	if(verbose.func){
		print("in retrieve", quote=FALSE)
		}
			
	lhs = lhs.steps[[currStep,1]]
	

# !! WARNING !!
#! this block of code is only for testing one Aphasia model. it should be removed/commented otherwise.

#	tmpChunk = chunks[[ret.que[1,1],1]]
#	if(verbose){
#		print("*** DEFAULT RETRIEVAL ***", quote=FALSE)
#		print(paste("Retrieving", tmpChunk[["name"]], "at: ", (currTime+ret.latency(ret.que[1,2])), " Activation:", ret.que[1,2]), quote=FALSE)
#	}
#	output.ret[currSim,currStage] <<- tmpChunk[["name"]]
#	output.time[currSim,currStage] <<- currTime + ret.latency(ret.que[1,2])
#	output.latency[currSim,currStage] <<- round(ret.latency(ret.que[1,2]), 4)*1000



	i=1
	while(i <= length(ret.que[,1])){
		currChunk = chunks[[ret.que[i,1],1]]


# if the lhs has a '=retrieval' request with at least one slot.
# this is to ensure that match remains '0' even when there is no '=retrieval' (i.e., when length(names(lhs)==0 in the next while loop)
		if(length(names(lhs)) > 0){
			match = 1	
			}
		else{
			match = 0
			}

		
		j=1
		while(j <= length(names(lhs))){
#print(paste("**** lhs names: ", lhs[[j]]))
			if(lhs[[j]] != currChunk[names(lhs)[j]]){
				match=0
				break
				}
			j=j+1
			}
			
		if(match == 1){
			if(ret.que[i,2] >= par.rt){
				add.time(ret.latency(ret.que[i,2]))

				if(verbose){
					print(paste("Retrieving", currChunk[["name"]], "at: ", currTime, " Activation:", ret.que[i,2]), quote=FALSE)
				}
				
				#! update no. of presentations of the chunk
				chunks[[ret.que[i,1],1]][["npres"]] <<- chunks[[ret.que[i,1],1]][["npres"]] + 1
				
				#! update retrieval history
				npres = chunks[[ret.que[i,1],1]][["npres"]] 
				ret.hist[ret.que[i,1], npres] <<- currTime
				
				output.ret[currSim,currStage] <<- currChunk[["name"]]
				output.time[currSim,currStage] <<- currTime
				output.latency[currSim,currStage] <<- round(ret.latency(ret.que[i,2]), 4)*1000
			}
			else{

				if(verbose){
					print("Activation below retrieval threshold!!", quote=FALSE)
				}
			}
			break
			}
		i=i+1
		}

	}


######################################################
# prepares a table of current retrieval cues with
# feature name and feature value as columns
######################################################

list.cues <- function(cues){
	if(verbose.func){
		print("in list.cues", quote=FALSE)
		}
		

	if(!is.null(cues)){
		nm = names(cues)
		
		if(!is.null(nm)){
			
			cues.list <<- array(, c(length(nm),3))

			for(j in c(1:length(nm))){
				cues.list[[j,1]] <<- nm[j]
				
				
				#! if the cue value is empty, the it has to be copied from the goal buffer
				#! slot but the 'type' cue value has to be always from the retrieval buffer
				
				if(cues[[nm[j]]] == ""){

					if(verbose){
print(paste("cue", j, ":", nm[j], "=", goal.chunk[[nm[j]]]), quote=FALSE)
}

					cues.list[[j,2]] <<- goal.chunk[[nm[j]]]
					}
				else{

					if(verbose){
		print(paste("cue", j, ":", nm[j], "=", cues[[nm[j]]]), quote=FALSE)
		}

					cues.list[[j,2]] <<- cues[[nm[j]]]
					}

				cues.list[[j,3]] <<- 1
				}
			}
		}
	}



######################################################
# builds a que with all retrieval target chunks
# also prepares 'fan' entry for each cue
######################################################

build.retQ <- function(){
	if(verbose.func){
		print("in build.retQ", quote=FALSE)
		}

	ind=c()
		
	i=1
	j=1
	currChunk = chunks[[i,1]]

	while(!is.null(currChunk)){
		
		if(currChunk[["type"]] == cues.list[[1,2]]){
			ind = c(ind, i)
			j=j+1
			}
		
		# prepare 'fan' entry for each cue
		for(l in c(2:length(cues.list[,2]))){
		# 1st entry in the list of cues is the 'type'

			for(k in c(1:length(currChunk))){
				if(cues.list[[l,2]] == currChunk[k]){
					cues.list[[l,3]] <<- as.numeric(cues.list[[l,3]])+1
				}
			}
		}
		
		i=i+1
		currChunk = chunks[[i,1]]
	}
	ret.que <<- array(c(ind,rep(0,length(ind))), c(length(ind),2))

}


######################################################
# computes spreading of activation
######################################################

compute.act.spreading <- function(){
	if(verbose.func){
		print("in compute.act.spreading", quote=FALSE)
		}

#temp.. comment out this:
#	W = 1/(length(cues.list[,2])+1)

#temp.. uncomment this:
	W = 1/(length(cues.list[,2]) - 1 + 1)
	#! length '- 1' to exclude 'type' feature
	#! length '+ 1' to exclude 'goal-cat' feature which is normally part of the goal buffer
	
	i=1
	while(i<=length(ret.que[,1])){
		currChunk = chunks[[ret.que[i,1],1]]
		
		j=2
		while(j <=length(cues.list[,2])){

			k=1
			while(k <=length(currChunk)){
				if(cues.list[[j,2]] == currChunk[k]){
					spread = (par.mas - log(as.numeric(cues.list[[j,3]])))
					act = W * spread
					
					if(verbose){
					print(paste("spreading act", act, "(W =", W, ", Sij =", spread, ") from", cues.list[[j,2]], "to chunk", currChunk[["name"]]), quote=FALSE)
					}
					
					ret.que[i,2] <<- ret.que[i,2] + act
					}
				k=k+1
				}
			j=j+1
			}
		i=i+1
		}
	}



######################################################
# computes base activation and adds noise for each
# chunk in the retrieval que
######################################################

compute.base.act <- function(){
	if(verbose.func){
		print("in compute.base.act", quote=FALSE)
		}

	i=1
	while(i<=length(ret.que[,1])){
		currChunk = chunks[[ret.que[i,1],1]]
		
#		ret.que[i,2] <<- ret.que[i,2] + Bi(as.numeric(currChunk["npres"]), (currTime-as.numeric(currChunk["ctime"]))) + noise(0, par.ans)
		ret.que[i,2] <<- ret.que[i,2] + Bi(ret.que[i,1]) + noise(0, par.ans)
		i=i+1
		}
	}


######################################################
# base level activation formula
######################################################

Bi <- function(chunkNo){
	if(verbose.func){
		print("in Bi", quote=FALSE)
		}
	
	ret = ret.hist[chunkNo,]

#! tj: duration since each retrieval
	ret = ifelse(ret==-1, 0, currTime-ret)
#! calculate: tj ^ -d
	ret = ifelse(ret==0, 0, ret^ (-1*par.bll))

	base = log(sum(ret))
	
	if(verbose){
	print(paste("Chunk:", chunks[[chunkNo,1]][["name"]]), quote=FALSE)
	print(paste("Base level: ", base), quote=FALSE)
	}

#	base=log(npres/(1-par.bll)) - (par.bll*log(lifeTime))
#	print(paste("Base level: ", base))

#	sq = c(1:npres)
#	base = log(sum((sq * lifeTime/npres) ^ (-1*par.bll)))
	
	return(base)
	}


######################################################
# logistic noise
######################################################

noise <- function(location, scale){
	if(verbose.func){
		print("in noise", quote=FALSE)
		}

	n = rlogis(1, location=location, scale=scale)
#	print(paste("Noise: ", n))
	return(n)
	
	#! location = mean of the logistic distribution
	#! scale = parameter for calculating variance of the logistic distribution
	
	}


######################################################
# retrieval time
######################################################

ret.latency <- function(act){
	return(par.lf*exp(-(act)))
	}


######################################################
# computes partial matching between cues
# and chunks
######################################################

compute.partial.match <- function(){
	if(verbose.func){
		print("in compute.partial.match", quote=FALSE)
		}

	i=1
	while(i <= length(ret.que[,1])){
		currChunk = chunks[[ret.que[i,1],1]]
		
		j=2
		while(j <= length(cues.list[,2])){
			if(!is.null(currChunk[[cues.list[[j,1]]]])){
			if(cues.list[[j,2]] == currChunk[[cues.list[[j,1]]]]){
				
				if(verbose){
print(paste("similarity between", cues.list[[j,2]], "and", currChunk[[cues.list[[j,1]]]], "in chunk", currChunk[["name"]], "is", par.ms), quote=FALSE)
}
				ret.que[i,2] <<- ret.que[i,2] + (par.mp * par.ms)
				}
			else{

				if(verbose){
print(paste("similarity between", cues.list[[j,2]], "and", currChunk[[cues.list[[j,1]]]], "in chunk", currChunk[["name"]], "is", par.md), quote=FALSE)
}
				ret.que[i,2] <<- ret.que[i,2] + (par.mp * par.md)
				}
			}
			j=j+1
			}
		i=i+1
		}

	}


######################################################
# sorts the retrieval que on the activations of 
# the chunks
######################################################

sort.retQ <- function(){
	tmp = cbind(ret.que[order(ret.que[,2], decreasing=TRUE),1], ret.que[order(ret.que[,2], decreasing=TRUE),2])
	ret.que <<- tmp
	}




######################################################
# Run the model
######################################################

run.model <- function(model, dir, run){
#run.model <- function(model, dir){

source(model)

#sink(paste(dir,"/output-", run, ".txt", sep=""))

output.ret <<- array(, c(sims,stages))
output.time <<- array(, c(sims,stages))
output.latency <<- array(, c(sims,stages))
output.prod <<- array(, c(sims,stages))

params = as.data.frame(rbind(c(par.mas, par.bll, par.ans, par.lf, par.rt, par.mp, par.ms, par.md, par.dat, par.egs)))
colnames(params)=c("mas", "decay", "act.noise", "latency", "ret.thr", "match.sc", "max.sim", "max.diff", "prod.tm", "prod.noise")

	if(verbose){
print(params, quote=FALSE)
}

for(s in c(1:sims)){

	currSim <<- s

	source("../R-ACT-R/global-var.R")
	
#add.chunk(pic1)
#add.chunk(pic2)

	if(verbose){
	print("",quote=FALSE)
	print("",quote=FALSE)
	print("=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-", quote=FALSE)
	print(paste("                    Simulation", s, "                    "), quote=FALSE)
	print("=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-", quote=FALSE)
      }

for(i in c(1:stages)){

#######################################################
# !! WARNING !!
# this block of code is only for the aphasia model
# it should be commented out otherwise

# create 2 picture chunks after NP1 is processed.
# both chunks are created in the same production.
# this is similar to creating syntactic chunks with
# buffers like NPb, VPb etc.

#	if(i==3){
#		add.time(par.dat)

#		if(verbose){
#			print(paste("------ Adding picture chunks -------"), quote=FALSE)
#			}
#		pic1[["ctime"]] = currTime
#		#pic1[["npres"]] = 1
#		add.chunk(pic1)

#		pic2[["ctime"]] = currTime
#		#pic2[["npres"]] = 1
#		add.chunk(pic2)
#		}
########################################################



	if(verbose){
	print("",quote=FALSE)
	print(paste("===============  ", "Stage ", i, ":", input.stages[i], "  ==============="), quote=FALSE)
	}


#! time for: 'Find-Location-Of-Next-Word'
	add.time(par.dat)
	if(verbose){
		print(paste("Production fired:", "Find-Location-Of-Next-Word"), quote=FALSE)
	}

#! time for: 'Attend-To-Next-Word'
	add.time(par.dat)
	if(verbose){
		print(paste("Production fired:", "Attend-To-Next-Word"), quote=FALSE)
	}

#! time for visual encoding of the word
	add.time(0.085)
	if(verbose){
		print("Visual encoding done", quote=FALSE)
	}

#! time for: lexical-retrieval-request
	add.time(par.dat)
	if(verbose){
		print(paste("Production fired:", " Lexical-Retrieval-Request-", input.stages[i], sep=""), quote=FALSE)
	}


	currStage <<- i
	currStep <<- conflict.resolution(i)
	
	check.goal()
	
	ret = ret.steps[[currStep,1]]
	
	if(verbose){
		print(paste("Production selected:", prods.names[currStep]), quote=FALSE)
	}

	output.prod[currSim,currStage] <<- prods.names[currStep]
	
#! time for: Set-Retrieval-Cues
	add.time(par.dat)
	if(verbose){
		print(paste("Production fired:", " Set-Retrieval-Cues-", input.stages[i], sep=""), quote=FALSE)
	}

	if(length(names(ret))>0){
		if(verbose){ print("------- retrieval --------", quote=FALSE)}

		retrieval.request(ret)
		retrieve()
	}


#! check if there is any new chunk creation
	if((length(names(vp.steps[[currStep,1]])) > 0)||
		(length(names(np.steps[[currStep,1]])) > 0)||
		(length(names(dp.steps[[currStep,1]])) > 0)||
		(length(names(ip.steps[[currStep,1]])) > 0)||
		(length(names(cp.steps[[currStep,1]])) > 0)||
		(length(names(adjp.steps[[currStep,1]])) > 0)||
		(length(names(pp.steps[[currStep,1]])) > 0))
	{
		newChunkCreation = TRUE
		}
	else{
		newChunkCreation = FALSE
		}

	
	if(length(names(lhs.steps[[currStep,1]])) > 0 || newChunkCreation){
	#! if there is a productions that has a =retrieval or that creates new chunks then add time
		if(verbose){
			print(paste("Production fired:", " Attach-", input.stages[i], sep=""), quote=FALSE)
			}
		add.time(par.dat)
		}


	if(newChunkCreation){
		if(verbose){
			print("------- creation --------", quote=FALSE)
			}

		#! create new chunks in various buffers (VPb, NPb etc.)
		cre = vp.steps[[currStep,1]]
		if(length(names(cre)) > 0){
			cre[["ctime"]] = currTime
			cre[["npres"]] = 1
			add.chunk(cre)
			}
		
		cre = np.steps[[currStep,1]]
		if(length(names(cre)) > 0){
			cre[["ctime"]] = currTime
			cre[["npres"]] = 1
			add.chunk(cre)
			}
		
		cre = dp.steps[[currStep,1]]
		if(length(names(cre)) > 0){
			cre[["ctime"]] = currTime
			cre[["npres"]] = 1
			add.chunk(cre)
			}
		
		cre = ip.steps[[currStep,1]]
		if(length(names(cre)) > 0){
			cre[["ctime"]] = currTime
			cre[["npres"]] = 1
			add.chunk(cre)
			}
		
		cre = cp.steps[[currStep,1]]
		if(length(names(cre)) > 0){
			cre[["ctime"]] = currTime
			cre[["npres"]] = 1
			add.chunk(cre)
			}
		
		cre = adjp.steps[[currStep,1]]
		if(length(names(cre)) > 0){
			cre[["ctime"]] = currTime
			cre[["npres"]] = 1
			add.chunk(cre)
			}
		
		cre = pp.steps[[currStep,1]]
		if(length(names(cre)) > 0){
			cre[["ctime"]] = currTime
			cre[["npres"]] = 1
			add.chunk(cre)
			}
		
		}
		
	}

	chunks <<- array(list(NULL), c(maxChunks,1))

	}
	dt.ret <<- as.data.frame(output.ret)
	dt.time <<- as.data.frame(output.time)
	dt.latency <<- as.data.frame(output.latency)
#	dt.prod <<- as.data.frame(output.prod)

	colnames(dt.ret) <<- input.stages
	colnames(dt.time) <<- input.stages
	colnames(dt.latency) <<- input.stages
#	colnames(dt.prod) <<- input.stages

	write.table(params, paste(dir,"/params-", run, ".txt", sep=""), quote=FALSE, row.names=FALSE)
	write.table(dt.ret, paste(dir, "/retrievals-", run, ".txt", sep=""), quote=FALSE, row.names=FALSE)
	write.table(dt.latency, paste(dir, "/ret-latency-", run, ".txt", sep=""), quote=FALSE, row.names=FALSE)

#	write.table(dt.time, paste(dir, "/ret-times-", run, ".txt", sep=""), quote=FALSE, row.names=FALSE)
#	write.table(dt.prod, paste(dir,"/productions-", run, ".txt", sep=""), quote=FALSE, row.names=FALSE)


#	sink(file=NULL)
	}



## TEMP 



